from SimPy.Simulation import *
from Schedule import *
from itertools import *
import string
import time

#from OrderedSchedule import * deprecated
#from RunAllAndInterrupt import * deprecated
from Interrupt0 import *
from Interrupt1 import *
from Interrupt2 import *
#from Static0 import *
#from Static1 import *
#from Static2 import *

from FIFOMistral import *
#from RRMistral import *

import scipy.stats as sp
import numpy as np

import csv
import os

DIR_PATH = "../logs/" #"/home/jacopo/Scrivania/work/cp_2011/logs/"
FOLDS_PATH = "../folds/" #"/home/jacopo/Scrivania/work/cp_2011/folds/"
NCLIENTS = 1 # number of processors

################################################################
# Schedules to test
################################################################
def concatenate(lists):
        return reduce( list.__add__, lists)

def create_orders_aux(schedule,schedule_name_pre,schedule_name_post,funct,funct_names, other_parameter):
    def aux(pairs):
        functs, names = zip(*pairs)
        return [
            schedule_name_pre + string.join(names, "<") + schedule_name_post,
            schedule,
            [compose_sort(list(functs)),
            other_parameter
            ]]
            
    per = permutations(zip(funct,funct_names))
    return map(aux,per)

def create_orders(schedule,schedule_name_pre,schedule_name_post,funct,funct_names, other_parameter):
    def allsubsets(ls):
        return concatenate(map(
            lambda x: list(combinations(ls,x)),
            range(1,len(ls)+1)))

    return concatenate(map(
        lambda x : create_orders_aux(schedule,schedule_name_pre,schedule_name_post,x[0],x[1], other_parameter),
        zip(allsubsets(funct), allsubsets(funct_names))
        ))



baselines = []

baselines.append([
    "FIFO_SPT_oracle",
    Interrupt0,
    [compose_sort([fast_key]),
    lambda x: False]
    ])

baselines.append([
    "FIFO_random_oracle",
    Interrupt0,
    [random.shuffle,
    lambda x: False
    ]])

baselines.append([
    "FIFO_SPT_fs",
    Interrupt1,
    [compose_sort([fast_key]),
    lambda x: False]
    ])

baselines.append([
    "FIFO_random_fs",
    Interrupt1,
    [random.shuffle,
    lambda x: False]
    ])

baselines.append([
    "FIFO_SPT_random",
    Interrupt2,
    [compose_sort([fast_key]),
    lambda x: False]
    ])

baselines.append([
    "FIFO_random_random",
    Interrupt2,
    [random.shuffle,
    lambda x: False]
    ])

baselines.append([
    "FIFO_mistral_mistral",
    FIFOMistral,
    [compose_sort([mistral_key])]
    ])
    
baselines.append([
    "FIFO_3C<max<avg<stdv_oracle",
    Interrupt0,
    [compose_sort([threec_key, max_key, avg_key, stdv_key]),
    lambda x: False]
    ])

baselines.append([
    "FIFO_3C<max<avg<stdv_random",
    Interrupt2,
    [compose_sort([threec_key, max_key, avg_key, stdv_key]),
    lambda x: False]
    ])

baselines.append([
    "FIFO_3C_oracle",
    Interrupt0,
    [compose_sort([threec_key]),
    lambda x: False]
    ])

simulations = []

simulations.append([
    "FIFO_3C<max<avg<stdv_fs",
    Interrupt1,
    [compose_sort([threec_key, max_key, avg_key, stdv_key]),
    lambda x: False]
    ])

simulations.extend(
    create_orders(
        Interrupt1,
        "FIFO_",
        "_fs",
        [threec_key, avg_key, stdv_key, max_key],
        ["3C", "avg", "stdv", "max"],
        lambda x: False
        )
    )


"""
simulations.extend(
    create_orders(
        Interrupt1,
        "FIFO_",
        "_fs",
        [invert(threec_key), invert(avg_key), stdv_key, invert(max_key)],
        ["{\\bar 3C}", "{\\bar avg}", "stdv", "{\\bar max}"],
        lambda x: False
        )
    )

simulations.extend(
    create_orders(
        Interrupt1,
        "FIFO_",
        "_fs",
        [invert(threec_key), invert(avg_key), invert(stdv_key), invert(max_key)],
        ["{\\bar 3C}", "{\\bar avg}", "{\\bar stdv}", "{\\bar max}"],
        lambda x: False
        )
    )
"""

"""
simulations.append([
    "RR_oracle_oracle",
    Static0,
    [compose_sort([fast_key]),
    NCLIENTS]
    ])

simulations.append([
    "interrupt_1",
    Interrupt1,
    [compose_sort([threec_key, avg_key, stdv_key, max_key]),
    selection_1]
    ])

simulations.append([
    "shortest_first_random_interrupt",
    Interrupt1,
    [compose_sort([threec_key, avg_key, stdv_key, max_key]),
    lambda x: True]
    ])
"""

simulations.sort(key= lambda x: x[0])
last = simulations[-1]
for i in range(len(simulations)-2, -1, -1):
    if last[0] == simulations[i][0]:
        del simulations[i]
    else:
        last = simulations[i]

################################################################
# Functions definitions
################################################################

def simulate_fold(file_name, output_file, (id, schedule, args)):
    initialize()
    s = schedule(file_name,*args)
    s.create_jobs()
    for i in range(NCLIENTS):
        c = Client(str(i),s)
        activate(c,c.start())
    simulate(until=10000000.0)

    jobs = len(s.instances)
    makespan = max([i for i in s.finishing_times.values()])
    sum_running_times = sum([i for i in s.running_times.values()])
    sum_solving_times = sum([i for i in s.solving_times.values()])
    output_file.writerow([
    sum_solving_times/jobs,
    makespan,
    np.median(s.solving_times.values()),
    sum_running_times/NCLIENTS,
    (makespan-sum_running_times/NCLIENTS)/makespan,
    len(s.solving_times),
    jobs,
    s.interrupts_num,
    s.solver_calls["mistral"],
    s.solver_calls["choco"],
    s.solver_calls["abscon"],
    ])

def write_first_line(file):
    file.writerow([
        "average solving time",
        "makespan",
        "median solving time"
        "average processor use",
        "% unused",
        "instances solved",
        "total instances",
        "interrupts number",
        "mistralCalls",
        "chocoCalls",
        "absconCalls"])

def main():
    files = filter( lambda x: x.startswith("cv"),os.listdir(FOLDS_PATH))
    files.sort()

    out_files_names = {}
    out_files_pointers = {}
    cvs_writer = {}

    for i in simulations:
        out_files_names[i[0]] = DIR_PATH + i[0] + ".csv"
        out_files_pointers[i[0]] = open(out_files_names[i[0]], 'wb')
        cvs_writer[i[0]] = csv.writer(out_files_pointers[i[0]])
        write_first_line(cvs_writer[i[0]])

    for i in baselines:
        out_files_names[i[0]] = DIR_PATH + i[0] + ".csv"
        out_files_pointers[i[0]] = open(out_files_names[i[0]], 'wb')
        cvs_writer[i[0]] = csv.writer(out_files_pointers[i[0]])
        write_first_line(cvs_writer[i[0]])
        
    for i in files:
        print "Simulation of %s" %(i)
        for j in simulations:
            #print "todel simultate %s" %(j[0])
            simulate_fold(FOLDS_PATH + i, cvs_writer[j[0]], j)
        for j in baselines:
            #print "todel simultate %s" %(j[0])
            simulate_fold(FOLDS_PATH + i, cvs_writer[j[0]], j)
        
    print "Simulation terminated"

    for i in simulations:
        out_files_pointers[i[0]].close()

    for i in baselines:
        out_files_pointers[i[0]].close()

    print "\nStatistics"

    averages = {}
    medians = {}
    mkspan = {}
    mistral_calls = {}
    abscon_calls = {}
    choco_calls = {}

    for i in simulations:
        averages[i[0]] = []
        mkspan[i[0]] = []
        medians[i[0]] = []
        mistral_calls[i[0]] = 0
        abscon_calls[i[0]] = 0
        choco_calls[i[0]] = 0
        reader = csv.reader(open(out_files_names[i[0]], "rb"))
        row_number = 0
        for row in reader:
            if row_number != 0:
                averages[i[0]].append(float(row[0]))
                mkspan[i[0]].append(float(row[1]))
                medians[i[0]].append(float(row[2]))
                mistral_calls[i[0]] += int(row[-3])
                abscon_calls[i[0]] += int(row[-1])
                choco_calls[i[0]] += int(row[-2])
            else:
                row_number += 1

    for i in baselines:
        averages[i[0]] = []
        mkspan[i[0]] = []
        medians[i[0]] = []
        mistral_calls[i[0]] = 0
        abscon_calls[i[0]] = 0
        choco_calls[i[0]] = 0
        reader = csv.reader(open(out_files_names[i[0]], "rb"))
        row_number = 0
        for row in reader:
            if row_number != 0:
                averages[i[0]].append(float(row[0]))
                mkspan[i[0]].append(float(row[1]))
                medians[i[0]].append(float(row[2]))
                mistral_calls[i[0]] += int(row[-3])
                abscon_calls[i[0]] += int(row[-1])
                choco_calls[i[0]] += int(row[-2])
            else:
                row_number += 1

    average = {}
    median_avg = {}
    std = {}
    avg_median = {}
    mkspan_average = {}
    mkspan_median = {}

    for i in simulations:
        average[i[0]] = np.mean(averages[i[0]])
        median_avg[i[0]] = np.median(averages[i[0]])
        avg_median[i[0]] = np.mean(medians[i[0]])
        std[i[0]] = np.std(averages[i[0]])
        mkspan_average[i[0]] = np.mean(mkspan[i[0]])
        #mkspan_average[i[0]] = sum(mkspan[i[0]]) / len(mkspan[i[0]])
        mkspan_median[i[0]] = np.median(mkspan[i[0]])

    best_simulation_avg = min(average.keys(), key=lambda x:average[x])
    best_simulation_mkspan = min(mkspan_average.keys(), key=lambda x:mkspan_average[x])

    print "best for minimizing averages: %s" % best_simulation_avg
    print "best for minimizing makespan: %s" % best_simulation_mkspan

    for i in baselines:
        average[i[0]] = np.mean(averages[i[0]])
        median_avg[i[0]] = np.median(averages[i[0]])
        avg_median[i[0]] = np.mean(medians[i[0]])
        std[i[0]] = np.std(averages[i[0]])
        mkspan_average[i[0]] = np.mean(mkspan[i[0]])
        #mkspan_average[i[0]] = sum(mkspan[i[0]]) / len(mkspan[i[0]])
        mkspan_median[i[0]] = np.median(mkspan[i[0]])

    print "order , average, median, avgMedian, std, mkspan average, mkspan median, significanceBestAvg, significanceBestMkspan, mistral, choco, abscon"

    simulations.extend(baselines)
    for i in simulations:
        (t,p_1) = sp.ttest_rel(averages[i[0]], averages[best_simulation_avg])
        (t,p_2) = sp.ttest_rel(mkspan[i[0]], mkspan[best_simulation_mkspan])
        print "%s," % (i[0]),
        print "%s," % (average[i[0]]),
        print "%s," % (avg_median[i[0]]),
        print "%s," % (median_avg[i[0]]),
        print "%s," % (std[i[0]]),
        print "%s," % (mkspan_average[i[0]]),
        print "%s," % (mkspan_median[i[0]]),
        if p_1 < 0.05:
            print "1,",
        else:
            print "0,",
        if p_2 < 0.05:
            print "1,",
        else:
            print "0,",
        print "%s," % (mistral_calls[i[0]]),
        print "%s," % (choco_calls[i[0]]),
        print "%s" % (abscon_calls[i[0]])

globaltime = time.time()
main()
print "Time spend for these tests: %s" % (time.time() - globaltime)
